%%%%%%%%%%%%%%%%%%%%%%
%                    %
%  STATE INVARIANTS  %
%                    %
%%%%%%%%%%%%%%%%%%%%%%

configInv(Config,M) :-
  Config = [_,_,_,_,_,_,MinPreservedLogSize,AlarmThresholdSize] &

  AlarmThresholdSize < MinPreservedLogSize &
  MinPreservedLogSize =< M & maxSupportedLogSize(M).

keyStoreInv(KeyStore,[M1,M2]) :- 
  KeyStore = [IssuerKey,OwnName] &

  (OwnName = {}
   or
   the(OwnName,M1)  & ncomp({[M1,M1]},IssuerKey,{}) % [M1,_] in IssuerKey
%   M1 in M2 &
%    the(OwnName,M1) &
%    dom(IssuerKey,M2)
  ).

adminInv(Admin,X) :-
  Admin = [RolePresent,AvailableOps,CurrentAdminOp] &

  (RolePresent neq {} or AvailableOps = {}) &
  (RolePresent = {} 
   or
   RolePresent neq {} & not_the(RolePresent,guard) 
   or
   AvailableOps = {overrideLock}
  ) &
  (RolePresent = {}
   or 
   RolePresent neq {} & not_the(RolePresent,auditManager)
   or
   AvailableOps = {archiveLog}
  ) &
  (RolePresent = {}
   or 
   RolePresent neq {} & not_the(RolePresent,securityOfficer)
   or 
   AvailableOps = {updateConfigData,shutdownOp}
  ) &
  (CurrentAdminOp = {}
   or 
   the(CurrentAdminOp,X) & X in AvailableOps & RolePresent neq {}
  ).

doorLatchAlarmInv(DoorLatchAlarm) :-
  DoorLatchAlarm =
    [CurrentTime,CurrentDoor,CurrentLatch,DoorAlarm,LatchTimeout,AlarmTimeout] &

  (CurrentLatch neq locked or CurrentTime >= LatchTimeout) &
  (CurrentTime < LatchTimeout or CurrentLatch = locked) &
  (DoorAlarm neq alarming
   or 
   CurrentDoor = open & CurrentLatch = locked & CurrentTime >= AlarmTimeout
  ) &
  (CurrentDoor neq open
   or 
   CurrentLatch neq locked
   or
   CurrentTime < AlarmTimeout 
   or
   DoorAlarm = alarming
  ).

idStationInv01(IDStation) :-
  IDStation = [_,_,_,DoorLatchAlarm,_,_,_,_,_,_,_,_,_,_] &
  
  doorLatchAlarmInv(DoorLatchAlarm).

idStationInv02(IDStation,M) :-
  IDStation = [_,_,_,_,_,_,Config,_,_,_,_,_,_,_] &

  configInv(Config,M).

idStationInv03(IDStation,[M1,M2]) :-
  IDStation = [_,_,_,_,_,_,_,_,KeyStore,_,_,_,_,_] &

  keyStoreInv(KeyStore,[M1,M2]).

idStationInv04(IDStation,X) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,_,_,_] &

  adminInv(Admin,X).

idStationInv05(IDStation,T) :-
  IDStation = [UserToken,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
    UserToken = [CurrentUserToken,_] &
    Internal = [Status,_,_] &

  (Status nin {gotFinger, waitingFinger, waitingUpdateToken, waitingEntry}
   or 
   validToken(T) & goodT(T) = CurrentUserToken
   or
   tokenWithValidAuth(T) & goodT(T) = CurrentUserToken
  ).

idStationInv06(IDStation,T) :-
  IDStation = [_,AdminToken,_,_,_,_,_,_,_,Admin,_,_,_,_] &
    Admin = [RolePresent,_,_] &
    AdminToken = [CurrentAdminToken,_] &

  (RolePresent = {} 
   or
   tokenWithValidAuth(T) & goodT(T) = CurrentAdminToken
  ).

idStationInv07(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,KeyStore,_,_,Internal,_,_] &
    KeyStore = [_,OwnName] &
    Internal = [_,EnclaveStatus,_] &

  (EnclaveStatus in {notEnrolled, waitingEnrol, waitingEndEnrol}
   or 
   OwnName neq {}
  ).

idStationInv08(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
    Internal = [_,EnclaveStatus,_] &
    Admin = [_,_,CurrentAdminOp] &

  (EnclaveStatus nin {waitingStartAdminOp, waitingFinishAdminOp}
   or
   CurrentAdminOp neq {}) &
  (CurrentAdminOp = {}
   or
   EnclaveStatus in {waitingStartAdminOp, waitingFinishAdminOp}
  ).

idStationInv09(IDStation,X) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
    Admin = [_,_,CurrentAdminOp] &
    Internal = [_,EnclaveStatus,_] &

  (CurrentAdminOp = {} 
   or
   the(CurrentAdminOp,X) & X nin {shutdownOp,overrideLock}
   or
   EnclaveStatus = waitingStartAdminOp
  ).

idStationInv10(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
    Internal = [_,EnclaveStatus,_] &
    Admin = [RolePresent,_,_] &

  (EnclaveStatus neq gotAdminToken
   or 
   RolePresent = {}
  ).

idStationInv11(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,Stats,_,_,_,_,_,CurrentScreen] &
  CurrentScreen = [ScreenStats,_,_] &

  ScreenStats = displayStats(Stats).

idStationInv12(IDStation) :-
  IDStation = [_,_,_,_,_,_,Config,_,_,_,_,_,_,CurrentScreen] &
    CurrentScreen = [_,_,ScreenConfig] &

  ScreenConfig = displayConfigData(Config).

idStationInv13(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,KeyStore,_,_,_,_,_] &
  KeyStore = [IssuerKey,_] &

  pfun(IssuerKey).

% negations

not_configInv(Config,M) :-
  Config = [_,_,_,_,_,_,MinPreservedLogSize,AlarmThresholdSize] &

  (AlarmThresholdSize >= MinPreservedLogSize
   or
   MinPreservedLogSize > M & maxSupportedLogSize(M)
  ).

not_keyStoreInv(KeyStore,[M1,M2]) :- 
  KeyStore = [IssuerKey,OwnName] &

%  OwnName neq {} & 
  the(OwnName,M1) & comp({[M1,M1]},IssuerKey,{}).
%  M1 nin M2 &
%    the(OwnName,M1) &
%    dom(IssuerKey,M2).

not_adminInv(Admin,X) :-
  Admin = [RolePresent,AvailableOps,CurrentAdminOp] &

  ((RolePresent = {} & AvailableOps neq {})
   or
   (RolePresent neq {} & 
    the(RolePresent,guard) & 
    AvailableOps neq {overrideLock})
   or
   (RolePresent neq {} & 
    the(RolePresent,auditManager) &
    AvailableOps neq {archiveLog})
   or
   (RolePresent neq {} & 
    the(RolePresent,securityOfficer) & 
    AvailableOps neq {updateConfigData,shutdownOp})
   or
   (CurrentAdminOp neq {} & 
    (the(CurrentAdminOp,X) & X nin AvailableOps
     or
     RolePresent = {}
    )
   )
  ).

not_doorLatchAlarmInv(DoorLatchAlarm) :-
  DoorLatchAlarm =
    [CurrentTime,CurrentDoor,CurrentLatch,DoorAlarm,LatchTimeout,AlarmTimeout] &

  ((CurrentLatch = locked & CurrentTime < LatchTimeout)
   or
   (CurrentTime >= LatchTimeout & CurrentLatch neq locked)
   or
   DoorAlarm = alarming &
   (CurrentDoor neq open
    or
    CurrentLatch neq locked
    or CurrentTime < AlarmTimeout
   ) 
   or
   (CurrentDoor = open & 
    CurrentLatch = locked & 
    CurrentTime >= AlarmTimeout & 
    DoorAlarm neq alarming
   )
  ).

not_idStationInv01(IDStation) :-
  IDStation = [_,_,_,DoorLatchAlarm,_,_,_,_,_,_,_,_,_,_] &
  
  not_doorLatchAlarmInv(DoorLatchAlarm).

not_idStationInv02(IDStation,M) :-
  IDStation = [_,_,_,_,_,_,Config,_,_,_,_,_,_,_] &

  not_configInv(Config,M).

not_idStationInv03(IDStation,[M1,M2]) :-
  IDStation = [_,_,_,_,_,_,_,_,KeyStore,_,_,_,_,_] &

  not_keyStoreInv(KeyStore,[M1,M2]).

not_idStationInv04(IDStation,X) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,_,_,_] &

  not_adminInv(Admin,X).

not_idStationInv05(IDStation,T) :-
  IDStation = [UserToken,_,_,_,_,_,_,_,_,_,_,Internal,_,_] &
    Internal = [Status,_,_] &
    UserToken = [CurrentUserToken,_] &

  Status in {gotFinger, waitingFinger, waitingUpdateToken, waitingEntry} &
  (goodT(T) = CurrentUserToken &
   not_validToken(T) &
   not_tokenWithValidAuth(T)
   or
   CurrentUserToken in {noFP,badFP}
  ).

not_idStationInv06(IDStation,T) :-
  IDStation = [_,AdminToken,_,_,_,_,_,_,_,Admin,_,_,_,_] &
    Admin = [RolePresent,_,_] &
    AdminToken = [CurrentAdminToken,_] &

  RolePresent neq {} & 
  (goodT(T) = CurrentAdminToken & not_tokenWithValidAuth(T)
   or
   CurrentAdminToken in {noFP,badFP}
  ).

not_idStationInv07(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,KeyStore,_,_,Internal,_,_] &
    Internal = [_,EnclaveStatus,_] &
    KeyStore = [_,OwnName] &

  EnclaveStatus nin {notEnrolled, waitingEnrol, waitingEndEnrol} &
  OwnName = {}.

not_idStationInv08(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
    Internal = [_,EnclaveStatus,_] &
    Admin = [_,_,CurrentAdminOp] &

  (EnclaveStatus in {waitingStartAdminOp, waitingFinishAdminOp} &
   CurrentAdminOp = {}
   or
   CurrentAdminOp neq {} &
   EnclaveStatus nin {waitingStartAdminOp, waitingFinishAdminOp}
  ).

not_idStationInv09(IDStation,X) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
    Admin = [_,_,CurrentAdminOp] &
    Internal = [_,EnclaveStatus,_] &

  CurrentAdminOp neq {} &
  the(CurrentAdminOp,X) & X in {shutdownOp,overrideLock} &
  EnclaveStatus neq waitingStartAdminOp.

not_idStationInv10(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,_,Admin,_,Internal,_,_] &
    Internal = [_,EnclaveStatus,_] &
    Admin = [RolePresent,_,_] &

  EnclaveStatus = gotAdminToken &
  RolePresent neq {}.

not_idStationInv11(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,_,_,_,_,_,CurrentScreen] &
    CurrentScreen = [ScreenStats,_,_] &

  (ScreenStats in 
     {clear, welcomeAdmin, busy, removeAdminToken, closeDoor,
      requestAdminOp, doingOp, invalidRequest, invalidData,
      insertEnrolmentData, validatingEnrolmentData, enrolmentFailed,
      archiveFailed, insertBlankFloppy, insertConfigData}
   or
   ScreenStats = displayConfigData(_)
  ).

not_idStationInv12(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,_,_,_,_,_,CurrentScreen] &
    CurrentScreen = [_,_,ScreenConfig] &

  (ScreenConfig in 
     {clear, welcomeAdmin, busy, removeAdminToken, closeDoor,
      requestAdminOp, doingOp, invalidRequest, invalidData,
      insertEnrolmentData, validatingEnrolmentData, enrolmentFailed,
      archiveFailed, insertBlankFloppy, insertConfigData}
   or
   ScreenConfig = displayStats(_)
  ).

not_idStationInv13(IDStation) :-
  IDStation = [_,_,_,_,_,_,_,_,KeyStore,_,_,_,_,_] &
  KeyStore = [IssuerKey,_] &

  npfun(IssuerKey).


